from __future__ import annotations

import calendar
import math
from string import whitespace
from typing import TYPE_CHECKING, Any

import sqlglot as sg
import sqlglot.expressions as sge

import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
from ibis import util
from ibis.backends.sql.compilers.base import NULL, STAR, AggGen, SQLGlotCompiler
from ibis.backends.sql.datatypes import ClickHouseType
from ibis.backends.sql.dialects import ClickHouse

if TYPE_CHECKING:
    from collections.abc import Iterator, Mapping


class ClickhouseAggGen(AggGen):
    def aggregate(self, compiler, name, *args, where=None, order_by=()):
        if order_by:
            raise com.UnsupportedOperationError(
                "ordering of order-sensitive aggregations via `order_by` is "
                "not supported for this backend"
            )
        # Clickhouse aggregate functions all have filtering variants with a
        # `If` suffix (e.g. `SumIf` instead of `Sum`).
        if where is not None:
            name += "If"
            args += (where,)
        return compiler.f[name](*args)


class ClickHouseCompiler(SQLGlotCompiler):
    __slots__ = ()

    dialect = ClickHouse
    type_mapper = ClickHouseType

    agg = ClickhouseAggGen()

    supports_qualify = True

    UNSUPPORTED_OPS = (
        ops.RowID,
        ops.CumeDist,
        ops.PercentRank,
        ops.Time,
        ops.TimeDelta,
        ops.StringToTimestamp,
        ops.StringToDate,
        ops.StringToTime,
        ops.Levenshtein,
    )

    SIMPLE_OPS = {
        ops.All: "min",
        ops.Any: "max",
        ops.ApproxCountDistinct: "uniqHLL12",
        ops.ApproxMedian: "median",
        ops.Arbitrary: "any",
        ops.ArrayContains: "has",
        ops.ArrayFlatten: "arrayFlatten",
        ops.ArrayIntersect: "arrayIntersect",
        ops.ArrayPosition: "indexOf",
        ops.BitwiseAnd: "bitAnd",
        ops.BitwiseNot: "bitNot",
        ops.BitwiseOr: "bitOr",
        ops.BitwiseXor: "bitXor",
        ops.Capitalize: "initcap",
        ops.CountDistinct: "uniq",
        ops.Date: "toDate",
        ops.E: "e",
        ops.EndsWith: "endsWith",
        ops.ExtractAuthority: "netloc",
        ops.ExtractDay: "toDayOfMonth",
        ops.ExtractDayOfYear: "toDayOfYear",
        ops.ExtractEpochSeconds: "toRelativeSecondNum",
        ops.ExtractFragment: "fragment",
        ops.ExtractHost: "domain",
        ops.ExtractHour: "toHour",
        ops.ExtractMinute: "toMinute",
        ops.ExtractMonth: "toMonth",
        ops.ExtractPath: "path",
        ops.ExtractProtocol: "protocol",
        ops.ExtractQuarter: "toQuarter",
        ops.ExtractSecond: "toSecond",
        ops.ExtractWeekOfYear: "toISOWeek",
        ops.ExtractYear: "toYear",
        ops.ExtractIsoYear: "toISOYear",
        ops.IntegerRange: "range",
        ops.IsInf: "isInfinite",
        ops.IsNan: "isNaN",
        ops.IsNull: "isNull",
        ops.Ln: "log",
        ops.Log10: "log10",
        ops.MapKeys: "mapKeys",
        ops.MapLength: "length",
        ops.MapMerge: "mapUpdate",
        ops.MapValues: "mapValues",
        ops.Median: "quantileExactExclusive",
        ops.NotNull: "isNotNull",
        ops.NullIf: "nullIf",
        ops.RegexReplace: "replaceRegexpAll",
        ops.RowNumber: "row_number",
        ops.StartsWith: "startsWith",
        ops.StrRight: "right",
        ops.Strftime: "formatDateTime",
        ops.StringLength: "length",
        ops.StringReplace: "replaceAll",
        ops.TimestampNow: "now",
        ops.TypeOf: "toTypeName",
        ops.Unnest: "arrayJoin",
        ops.RandomUUID: "generateUUIDv4",
        ops.RandomScalar: "randCanonical",
    }

    @staticmethod
    def _minimize_spec(start, end, spec):
        if (
            start is None
            and isinstance(getattr(end, "value", None), ops.Literal)
            and end.value.value == 0
            and end.following
        ):
            return None
        return spec

    def visit_Cast(self, op, *, arg, to):
        _interval_cast_suffixes = {
            "s": "Second",
            "m": "Minute",
            "h": "Hour",
            "D": "Day",
            "W": "Week",
            "M": "Month",
            "Q": "Quarter",
            "Y": "Year",
        }

        if to.is_interval():
            suffix = _interval_cast_suffixes[to.unit.short]
            return self.f[f"toInterval{suffix}"](arg)

        result = self.cast(arg, to)
        if (timezone := getattr(to, "timezone", None)) is not None:
            return self.f.toTimeZone(result, timezone)
        return result

    def visit_TryCast(self, op, *, arg, to):
        return self.f.accurateCastOrNull(arg, self.type_mapper.to_string(to))

    def visit_ArrayIndex(self, op, *, arg, index):
        return arg[self.if_(index >= 0, index + 1, index)]

    def visit_ArrayRepeat(self, op, *, arg, times):
        param = sg.to_identifier("_")
        func = sge.Lambda(this=arg, expressions=[param])
        return self.f.arrayFlatten(self.f.arrayMap(func, self.f.range(times)))

    def visit_ArraySlice(self, op, *, arg, start, stop):
        start = self._add_parens(start)
        start_correct = self.if_(start < 0, start, start + 1)

        if stop is not None:
            stop = self._add_parens(stop)

            length = self.if_(
                stop < 0,
                stop,
                self.if_(
                    start < 0,
                    self.f.greatest(0, stop - (self.f.length(arg) + start)),
                    self.f.greatest(0, stop - start),
                ),
            )
            return self.f.arraySlice(arg, start_correct, length)
        else:
            return self.f.arraySlice(arg, start_correct)

    def visit_CountStar(self, op, *, where, arg):
        if where is not None:
            return self.f.countIf(where)
        return sge.Count(this=STAR)

    def _visit_quantile(self, func, arg, quantile, where):
        return sge.ParameterizedAgg(
            this=f"{func}If" if where is not None else func,
            expressions=util.promote_list(quantile),
            params=[arg, where] if where is not None else [arg],
        )

    def visit_Quantile(self, op, *, arg, quantile, where):
        return self._visit_quantile("quantile", arg, quantile, where)

    def visit_MultiQuantile(self, op, *, arg, quantile, where):
        return self._visit_quantile("quantiles", arg, quantile, where)

    def visit_ApproxQuantile(self, op, *, arg, quantile, where):
        return self._visit_quantile("quantileTDigest", arg, quantile, where)

    def visit_ApproxMultiQuantile(self, op, *, arg, quantile, where):
        return self._visit_quantile("quantilesTDigest", arg, quantile, where)

    def visit_Correlation(self, op, *, left, right, how, where):
        if how == "pop":
            raise ValueError(
                "ClickHouse only implements `sample` correlation coefficient"
            )
        return self.agg.corr(left, right, where=where)

    def visit_StringFind(self, op, *, arg, substr, start, end):
        if end is not None:
            raise com.UnsupportedOperationError(
                "String find doesn't support end argument"
            )

        if start is not None:
            return self.f.position(arg, substr, start)

        return self.f.position(arg, substr)

    def visit_RegexSearch(self, op, *, arg, pattern):
        return sge.RegexpLike(this=arg, expression=pattern)

    def visit_RegexExtract(self, op, *, arg, pattern, index):
        arg = self.cast(arg, dt.String(nullable=False))

        pattern = self.f.concat("(", pattern, ")")

        if index is None:
            index = 0

        index += 1

        then = self.f.extractGroups(arg, pattern)[index]

        return self.if_(self.f.notEmpty(then), then, NULL)

    def visit_FindInSet(self, op, *, needle, values):
        return self.f.indexOf(self.f.array(*values), needle)

    def visit_Sign(self, op, *, arg):
        """Workaround for missing sign function in older versions of clickhouse."""
        return self.f.intDivOrZero(arg, self.f.abs(arg))

    def visit_Hash(self, op, *, arg):
        return self.f.reinterpretAsInt64(self.f.sipHash64(arg))

    def visit_HashBytes(self, op, *, arg, how):
        supported_algorithms = {
            "md5": "MD5",
            "MD5": "MD5",
            "halfMD5": "halfMD5",
            "SHA1": "SHA1",
            "sha1": "SHA1",
            "SHA224": "SHA224",
            "sha224": "SHA224",
            "SHA256": "SHA256",
            "sha256": "SHA256",
            "intHash32": "intHash32",
            "intHash64": "intHash64",
            "cityHash64": "cityHash64",
            "sipHash64": "sipHash64",
            "sipHash128": "sipHash128",
        }
        if (funcname := supported_algorithms.get(how)) is None:
            raise com.UnsupportedOperationError(f"Unsupported hash algorithm {how}")

        return self.f[funcname](arg)

    def visit_IntervalFromInteger(self, op, *, arg, unit):
        dtype = op.dtype
        if dtype.unit.short in ("ms", "us", "ns"):
            raise com.UnsupportedOperationError(
                "Clickhouse doesn't support subsecond interval resolutions"
            )
        return super().visit_IntervalFromInteger(op, arg=arg, unit=unit)

    def visit_NonNullLiteral(self, op, *, value, dtype):
        if dtype.is_inet():
            v = str(value)
            return self.f.toIPv6(v) if ":" in v else self.f.toIPv4(v)
        elif dtype.is_decimal():
            precision = dtype.precision
            if precision is None or not 1 <= precision <= 76:
                raise NotImplementedError(
                    f"Unsupported precision. Supported values: [1 : 76]. Current value: {precision!r}"
                )

            if 1 <= precision <= 9:
                type_name = self.f.toDecimal32
            elif 10 <= precision <= 18:
                type_name = self.f.toDecimal64
            elif 19 <= precision <= 38:
                type_name = self.f.toDecimal128
            else:
                type_name = self.f.toDecimal256
            return type_name(value, dtype.scale)
        elif dtype.is_numeric():
            if not math.isfinite(value):
                return sge.Literal.number(str(value))
            return sge.convert(value)
        elif dtype.is_interval():
            if dtype.unit.short in ("ms", "us", "ns"):
                raise com.UnsupportedOperationError(
                    "Clickhouse doesn't support subsecond interval resolutions"
                )

            return sge.Interval(
                this=sge.convert(str(value)), unit=dtype.resolution.upper()
            )
        elif dtype.is_timestamp():
            funcname = "parseDateTime"

            if micros := value.microsecond:
                funcname += "64"

            funcname += "BestEffort"

            args = [value.isoformat()]

            if micros % 1000:
                args.append(6)
            elif micros // 1000:
                args.append(3)

            if (timezone := dtype.timezone) is not None:
                args.append(timezone)

            return self.f[funcname](*args)
        elif dtype.is_date():
            return self.f.toDate(value.isoformat())
        elif dtype.is_array():
            value_type = dtype.value_type
            values = [
                self.visit_Literal(
                    ops.Literal(v, dtype=value_type), value=v, dtype=value_type
                )
                for v in value
            ]
            return self.f.array(*values)
        elif dtype.is_map():
            value_type = dtype.value_type
            keys = []
            values = []

            for k, v in value.items():
                keys.append(sge.convert(k))
                values.append(
                    self.visit_Literal(
                        ops.Literal(v, dtype=value_type),
                        value=v,
                        dtype=value_type,
                    )
                )

            return self.f.map(self.f.array(*keys), self.f.array(*values))
        elif dtype.is_struct():
            fields = [
                self.visit_Literal(
                    ops.Literal(v, dtype=field_type), value=v, dtype=field_type
                )
                for field_type, v in zip(dtype.types, value.values())
            ]
            return self.f.tuple(*fields)
        else:
            return None

    def visit_TimestampFromUNIX(self, op, *, arg, unit):
        if (unit := unit.short) in {"ms", "us", "ns"}:
            raise com.UnsupportedOperationError(f"{unit!r} unit is not supported!")
        return self.f.toDateTime(arg)

    def visit_TimestampTruncate(self, op, *, arg, unit):
        if (short := unit.short) == "W":
            funcname = "toMonday"
        else:
            funcname = f"toStartOf{unit.singular.capitalize()}"

        func = self.f[funcname]

        if short in ("Y", "Q", "M", "W", "D"):
            # these units return `Date` so we have to cast back to the
            # corresponding Ibis type
            return self.cast(func(arg), op.dtype)
        elif short in ("s", "ms", "us", "ns"):
            return func(self.f.toDateTime64(arg, op.arg.dtype.scale or 0))
        else:
            assert short in ("h", "m"), short
            return func(arg)

    visit_TimeTruncate = visit_TimestampTruncate

    def visit_DateTruncate(self, op, *, arg, unit):
        if unit.short == "D":
            # no op because truncating a date to a date has no effect
            return arg
        elif unit.short == "W":
            func = "toMonday"
        else:
            func = f"toStartOf{unit.singular.capitalize()}"

        # no cast needed here because all of the allowed units return `Date`
        # values
        return self.f[func](arg)

    def visit_TimestampBucket(self, op, *, arg, interval, offset):
        if offset is not None:
            raise com.UnsupportedOperationError(
                "Timestamp bucket with offset is not supported"
            )

        return self.f.toStartOfInterval(arg, interval)

    def visit_DateFromYMD(self, op, *, year, month, day):
        return self.f.toDate(
            self.f.concat(
                self.f.toString(year),
                "-",
                self.f.leftPad(self.f.toString(month), 2, "0"),
                "-",
                self.f.leftPad(self.f.toString(day), 2, "0"),
            )
        )

    def visit_TimestampFromYMDHMS(
        self, op, *, year, month, day, hours, minutes, seconds, **_
    ):
        to_datetime = self.f.toDateTime(
            self.f.concat(
                self.f.toString(year),
                "-",
                self.f.leftPad(self.f.toString(month), 2, "0"),
                "-",
                self.f.leftPad(self.f.toString(day), 2, "0"),
                " ",
                self.f.leftPad(self.f.toString(hours), 2, "0"),
                ":",
                self.f.leftPad(self.f.toString(minutes), 2, "0"),
                ":",
                self.f.leftPad(self.f.toString(seconds), 2, "0"),
            )
        )
        if timezone := op.dtype.timezone:
            return self.f.toTimeZone(to_datetime, timezone)
        return to_datetime

    def visit_StringSplit(self, op, *, arg, delimiter):
        return self.f.splitByString(
            delimiter, self.cast(arg, dt.String(nullable=False))
        )

    def visit_GroupConcat(self, op, *, arg, sep, where, order_by):
        if order_by:
            raise com.UnsupportedOperationError(
                "ordering of order-sensitive aggregations via `order_by` is "
                "not supported for this backend"
            )
        call = self.agg.groupArray(arg, where=where)
        return self.if_(self.f.empty(call), NULL, self.f.arrayStringConcat(call, sep))

    def visit_Cot(self, op, *, arg):
        return 1.0 / self.f.tan(arg)

    def visit_StructColumn(self, op, *, values, **_):
        return self.f.tuple(*values)

    def visit_Clip(self, op, *, arg, lower, upper):
        if upper is not None:
            arg = self.if_(self.f.isNull(arg), NULL, self.f.least(upper, arg))

        if lower is not None:
            arg = self.if_(self.f.isNull(arg), NULL, self.f.greatest(lower, arg))

        return arg

    def visit_StructField(self, op, *, arg, field: str):
        arg_dtype = op.arg.dtype
        idx = arg_dtype.names.index(field)
        return self.cast(sge.Dot(this=arg, expression=sge.convert(idx + 1)), op.dtype)

    def visit_Repeat(self, op, *, arg, times):
        return self.f.repeat(arg, self.f.accurateCast(times, "UInt64"))

    def visit_StringContains(self, op, haystack, needle):
        return self.f.position(haystack, needle) > 0

    def visit_Strip(self, op, *, arg):
        return sge.Trim(
            this=arg, position="BOTH", expression=sge.Literal.string(whitespace)
        )

    def visit_LPad(self, op, *, arg, length, pad):
        return self.f.leftPadUTF8(
            arg, self.f.greatest(self.f.lengthUTF8(arg), length), pad
        )

    def visit_RPad(self, op, *, arg, length, pad):
        return self.f.rightPadUTF8(
            arg, self.f.greatest(self.f.lengthUTF8(arg), length), pad
        )

    def visit_DayOfWeekIndex(self, op, *, arg):
        weekdays = len(calendar.day_name)
        return (((self.f.toDayOfWeek(arg) - 1) % weekdays) + weekdays) % weekdays

    def visit_DayOfWeekName(self, op, *, arg):
        # ClickHouse 20 doesn't support dateName
        #
        # ClickHouse 21 supports dateName is broken for regexen:
        # https://github.com/ClickHouse/ClickHouse/issues/32777
        #
        # ClickHouses 20 and 21 also have a broken case statement hence the ifnull:
        # https://github.com/ClickHouse/ClickHouse/issues/32849
        #
        # We test against 20 in CI, so we implement day_of_week_name as follows
        days = calendar.day_name
        num_weekdays = len(days)
        base = (
            ((self.f.toDayOfWeek(arg) - 1) % num_weekdays) + num_weekdays
        ) % num_weekdays
        return sge.Case(
            this=base,
            ifs=list(map(self.if_, *zip(*enumerate(days)))),
            default=sge.convert(""),
        )

    def visit_Map(self, op, *, keys, values):
        # cast here to allow lookups of nullable columns
        return self.cast(self.f.tuple(keys, values), op.dtype)

    def visit_MapGet(self, op, *, arg, key, default):
        return self.if_(self.f.mapContains(arg, key), arg[key], default)

    def visit_ArrayConcat(self, op, *, arg):
        return self.f.arrayConcat(*arg)

    def visit_BitAndOrXor(self, op, *, arg, where):
        if not (dtype := op.arg.dtype).is_unsigned_integer():
            nbits = dtype.nbytes * 8
            arg = self.f[f"reinterpretAsUInt{nbits}"](arg)
        return self.agg[f"group{type(op).__name__}"](arg, where=where)

    visit_BitAnd = visit_BitOr = visit_BitXor = visit_BitAndOrXor

    def visit_StandardDevVariance(self, op, *, how, where, **kw):
        funcs = {
            ops.StandardDev: "stddev",
            ops.Variance: "var",
            ops.Covariance: "covar",
        }
        func = funcs[type(op)]
        variants = {"sample": f"{func}Samp", "pop": f"{func}Pop"}
        funcname = variants[how]
        return self.agg[funcname](*kw.values(), where=where)

    visit_StandardDev = visit_Variance = visit_Covariance = visit_StandardDevVariance

    def visit_ArrayDistinct(self, op, *, arg):
        null_element = self.if_(
            self.f.countEqual(arg, NULL) > 0, self.f.array(NULL), self.f.array()
        )
        return self.f.arrayConcat(self.f.arrayDistinct(arg), null_element)

    def visit_ExtractMicrosecond(self, op, *, arg):
        dtype = op.dtype
        return self.cast(
            self.f.toUnixTimestamp64Micro(self.cast(arg, op.arg.dtype.copy(scale=6)))
            % 1_000_000,
            dtype,
        )

    def visit_ExtractMillisecond(self, op, *, arg):
        dtype = op.dtype
        return self.cast(
            self.f.toUnixTimestamp64Milli(self.cast(arg, op.arg.dtype.copy(scale=3)))
            % 1_000,
            dtype,
        )

    def visit_LagLead(self, op, *, arg, offset, default):
        args = [arg]

        if default is not None:
            if offset is None:
                offset = 1

            args.append(offset)
            args.append(default)
        elif offset is not None:
            args.append(offset)

        func = self.f[f"{type(op).__name__.lower()}InFrame"]
        return func(*args)

    visit_Lag = visit_Lead = visit_LagLead

    def visit_ExtractFile(self, op, *, arg):
        return self.f.cutFragment(self.f.pathFull(arg))

    def visit_ExtractQuery(self, op, *, arg, key):
        if key is not None:
            return self.f.extractURLParameter(arg, key)
        else:
            return self.f.queryString(arg)

    def visit_ArrayStringJoin(self, op, *, arg, sep):
        return self.f.arrayStringConcat(arg, sep)

    def visit_ArrayMap(self, op, *, arg, param, body, index):
        expressions = [param]
        args = [arg]

        if index is not None:
            expressions.append(index)
            args.append(self.f.range(0, self.f.length(arg)))

        func = sge.Lambda(this=body, expressions=expressions)

        return self.f.arrayMap(func, *args)

    def visit_ArrayFilter(self, op, *, arg, param, body, index):
        expressions = [param]
        args = [arg]

        if index is not None:
            expressions.append(index)
            args.append(self.f.range(0, self.f.length(arg)))

        func = sge.Lambda(this=body, expressions=expressions)

        return self.f.arrayFilter(func, *args)

    def visit_ArrayRemove(self, op, *, arg, other):
        x = sg.to_identifier(util.gen_name("x"))
        should_keep_null = sg.and_(x.is_(NULL), other.is_(sg.not_(NULL)))
        cond = sg.or_(x.neq(other), should_keep_null)
        return self.f.arrayFilter(sge.Lambda(this=cond, expressions=[x]), arg)

    def visit_ArrayUnion(self, op, *, left, right):
        arg = self.f.arrayConcat(left, right)
        null_element = self.if_(
            self.f.countEqual(arg, NULL) > 0, self.f.array(NULL), self.f.array()
        )
        return self.f.arrayConcat(self.f.arrayDistinct(arg), null_element)

    def visit_ArrayZip(self, op: ops.ArrayZip, *, arg, **_: Any) -> str:
        return self.f.arrayZip(*arg)

    def visit_ArrayCollect(self, op, *, arg, where, order_by, include_null, distinct):
        if include_null:
            raise com.UnsupportedOperationError(
                "`include_null=True` is not supported by the clickhouse backend"
            )
        func = self.agg.groupUniqArray if distinct else self.agg.groupArray
        return func(arg, where=where, order_by=order_by)

    def visit_First(self, op, *, arg, where, order_by, include_null):
        if include_null:
            raise com.UnsupportedOperationError(
                "`include_null=True` is not supported by the clickhouse backend"
            )
        return self.agg.any(arg, where=where, order_by=order_by)

    def visit_Last(self, op, *, arg, where, order_by, include_null):
        if include_null:
            raise com.UnsupportedOperationError(
                "`include_null=True` is not supported by the clickhouse backend"
            )
        return self.agg.anyLast(arg, where=where, order_by=order_by)

    def visit_ArgMin(self, op, *, arg, key, where):
        return sge.Dot(
            this=self.agg.argMin(self.f.tuple(arg), key, where=where),
            expression=sge.convert(1),
        )

    def visit_ArgMax(self, op, *, arg, key, where):
        return sge.Dot(
            this=self.agg.argMax(self.f.tuple(arg), key, where=where),
            expression=sge.convert(1),
        )

    def visit_CountDistinctStar(
        self, op: ops.CountDistinctStar, *, where, **_: Any
    ) -> str:
        columns = self.f.tuple(*map(sg.column, op.arg.schema.names))

        if where is not None:
            return self.f.countDistinctIf(columns, where)
        else:
            return self.f.countDistinct(columns)

    def visit_TimestampRange(self, op, *, start, stop, step):
        unit = op.step.dtype.unit.name.lower()

        if not isinstance(op.step, ops.Literal):
            raise com.UnsupportedOperationError(
                "ClickHouse doesn't support non-literal step values"
            )

        step_value = op.step.value

        offset = sg.to_identifier("offset")

        this = self.f.dateAdd(self.v[unit], offset, start)
        func = sge.Lambda(this=this, expressions=[offset])

        if step_value == 0:
            return self.f.array()

        return self.f.arrayMap(
            func, self.f.range(0, self.f.timestampDiff(unit, start, stop), step_value)
        )

    def visit_RegexSplit(self, op, *, arg, pattern):
        return self.f.splitByRegexp(pattern, self.cast(arg, dt.String(nullable=False)))

    @staticmethod
    def _generate_groups(groups):
        return groups

    def visit_DropColumns(self, op, *, parent, columns_to_drop):
        quoted = self.quoted
        excludes = [sg.column(column, quoted=quoted) for column in columns_to_drop]
        star = sge.Star(**{"except": excludes})
        table = sg.to_identifier(parent.alias_or_name, quoted=quoted)
        column = sge.Column(this=star, table=table)
        return sg.select(column).from_(parent)

    def visit_TableUnnest(
        self,
        op,
        *,
        parent,
        column,
        column_name: str,
        offset: str | None,
        keep_empty: bool,
    ):
        quoted = self.quoted

        column_alias = sg.to_identifier(
            util.gen_name("table_unnest_column"), quoted=quoted
        )

        table = sg.to_identifier(parent.alias_or_name, quoted=quoted)

        selcols = []

        overlaps_with_parent = column_name in op.parent.schema
        computed_column = column_alias.as_(column_name, quoted=quoted)

        if offset is not None:
            if overlaps_with_parent:
                selcols.append(
                    sge.Column(this=sge.Star(replace=[computed_column]), table=table)
                )
            else:
                selcols.append(sge.Column(this=STAR, table=table))
                selcols.append(computed_column)

            offset = sg.to_identifier(offset, quoted=quoted)
            selcols.append(offset)
        elif overlaps_with_parent:
            selcols.append(
                sge.Column(this=sge.Star(replace=[computed_column]), table=table)
            )
        else:
            selcols.append(sge.Column(this=STAR, table=table))
            selcols.append(computed_column)

        select = (
            sg.select(*selcols)
            .from_(parent)
            .join(
                sge.Join(
                    this=column.as_(column_alias, quoted=quoted),
                    kind="ARRAY",
                    side=None if not keep_empty else "LEFT",
                )
            )
        )

        if offset is not None:
            param = sg.to_identifier(util.gen_name("arr_enum"))
            func = sge.Lambda(this=param - 1, expressions=[param])
            return select.join(
                self.f.arrayMap(func, self.f.arrayEnumerate(column)).as_(offset)
            )

        return select

    def _cleanup_names(
        self, exprs: Mapping[str, sge.Expression]
    ) -> Iterator[sge.Expression]:
        """Compose `_gen_valid_name` and `_dedup_name` to clean up names in projections.

        ClickHouse has a bug where this fails to find the final `"o"."a"` column:

        ```sql
        SELECT
          "o"."a"
        FROM (
          SELECT
            "w"."a"
          FROM "t" AS "s"
          INNER JOIN "t" AS "w"
          USING ("a")
        ) AS "o"
        ```

        Adding a redundant aliasing operation (`"w"."a" AS "a"`) helps
        ClickHouse.
        """
        quoted = self.quoted
        return (
            value.as_(self._gen_valid_name(name), quoted=quoted, copy=False)
            for name, value in exprs.items()
        )

    def _array_reduction(self, arg):
        x = sg.to_identifier("x", quoted=self.quoted)
        not_null = sge.Lambda(this=x.is_(sg.not_(NULL)), expressions=[x])
        return self.f.arrayFilter(not_null, arg)

    def visit_ArrayMin(self, op, *, arg):
        return self.f.arrayReduce("min", self._array_reduction(arg))

    visit_ArrayAll = visit_ArrayMin

    def visit_ArrayMax(self, op, *, arg):
        return self.f.arrayReduce("max", self._array_reduction(arg))

    visit_ArrayAny = visit_ArrayMax

    def visit_ArraySum(self, op, *, arg):
        return self.f.arrayReduce("sum", self._array_reduction(arg))

    def visit_ArrayMean(self, op, *, arg):
        return self.f.arrayReduce("avg", self._array_reduction(arg))

    def _promote_bitshift_inputs(self, *, op, left, right):
        # clickhouse is incredibly pedantic about types allowed in bit shifting
        #
        # e.g., a UInt8 cannot be bitshift by more than 8 bits, UInt16 by more
        # than 16, and so on.
        #
        # This is why something like Ibis is necessary so that people have just
        # _consistent_ things, let alone *nice* things.
        left_dtype = op.left.dtype
        right_dtype = op.right.dtype

        if left_dtype != right_dtype:
            promoted = dt.higher_precedence(left_dtype, right_dtype)
            return self.cast(left, promoted), self.cast(right, promoted)
        return left, right

    def visit_BitwiseLeftShift(self, op, *, left, right):
        return self.f.bitShiftLeft(
            *self._promote_bitshift_inputs(op=op, left=left, right=right)
        )

    def visit_BitwiseRightShift(self, op, *, left, right):
        return self.f.bitShiftRight(
            *self._promote_bitshift_inputs(op=op, left=left, right=right)
        )

    def visit_MapContains(self, op, *, arg, key):
        return self.if_(
            sg.or_(arg.is_(NULL), key.is_(NULL)), NULL, self.f.mapContains(arg, key)
        )


compiler = ClickHouseCompiler()



from __future__ import annotations

import pytest
from pytest import mark

import ibis
import ibis.common.exceptions as com
from ibis.conftest import IS_SPARK_REMOTE

pa = pytest.importorskip("pyarrow")
ds = pytest.importorskip("pyarrow.dataset")

pytestmark = [
    mark.notyet(
        ["databricks"],
        reason="Databricks does not support temporary tables, even though they allow the syntax",
    )
]


@mark.notimpl(["datafusion", "flink", "impala", "trino", "druid"])
@mark.notimpl(["exasol"], reason="Exasol does not support temporary tables")
@pytest.mark.never(
    ["risingwave"],
    raises=com.UnsupportedOperationError,
    reason="Feature is not yet implemented: CREATE TEMPORARY TABLE",
)
def test_persist_expression(backend, alltypes):
    non_persisted_table = alltypes.mutate(
        test_column=ibis.literal("calculation"), other_calc=ibis.literal("xyz")
    )
    persisted_table = non_persisted_table.cache()
    backend.assert_frame_equal(
        non_persisted_table.order_by("id").to_pandas(),
        persisted_table.order_by("id").to_pandas(),
    )


@mark.notimpl(["datafusion", "flink", "impala", "trino", "druid"])
@mark.notimpl(["exasol"], reason="Exasol does not support temporary tables")
@pytest.mark.never(
    ["risingwave"],
    raises=com.UnsupportedOperationError,
    reason="Feature is not yet implemented: CREATE TEMPORARY TABLE",
)
def test_persist_expression_contextmanager(backend, con, alltypes):
    non_cached_table = alltypes.mutate(
        test_column=ibis.literal("calculation"), other_column=ibis.literal("big calc")
    )
    with non_cached_table.cache() as cached_table:
        backend.assert_frame_equal(
            non_cached_table.order_by("id").to_pandas(),
            cached_table.order_by("id").to_pandas(),
        )
    assert non_cached_table.op() not in con._cache_op_to_entry


@mark.notimpl(["flink", "impala", "trino", "druid"])
@pytest.mark.never(
    ["risingwave"],
    raises=com.UnsupportedOperationError,
    reason="Feature is not yet implemented: CREATE TEMPORARY TABLE",
)
@mark.notimpl(["exasol"], reason="Exasol does not support temporary tables")
@mark.notyet(
    ["pyspark"],
    condition=not IS_SPARK_REMOTE,
    raises=AssertionError,
    reason=(
        "PySpark holds on to `cached_table` in the stack frame of an internal function. "
        "Caching works, but this test is probably too strict for PySpark. "
        "On the other hand, it's hard to write a test that makes assertions after weakref "
        "finalizers are called."
    ),
)
def test_persist_expression_multiple_refs(backend, con, alltypes):
    non_cached_table = alltypes.mutate(
        test_column=ibis.literal("calculation"), other_column=ibis.literal("big calc 2")
    )
    op = non_cached_table.op()
    cached_table = non_cached_table.cache()

    backend.assert_frame_equal(
        non_cached_table.order_by("id").to_pandas(),
        cached_table.order_by("id").to_pandas(),
        check_dtype=False,
    )

    name = cached_table.op().name
    nested_cached_table = non_cached_table.cache()

    # cached tables are identical and reusing the same op
    assert cached_table.op() is nested_cached_table.op()
    # table is cached
    assert op in con._cache_op_to_entry

    # deleting the first reference, leaves table in cache
    del nested_cached_table
    assert op in con._cache_op_to_entry

    # deleting the last reference, releases table from cache
    del cached_table
    assert op not in con._cache_op_to_entry

    # assert that table has been dropped
    assert name not in con.list_tables()


@mark.notimpl(["flink", "impala", "trino", "druid"])
@mark.notimpl(["exasol"], reason="Exasol does not support temporary tables")
@pytest.mark.never(
    ["risingwave"],
    raises=com.UnsupportedOperationError,
    reason="Feature is not yet implemented: CREATE TEMPORARY TABLE",
)
def test_persist_expression_repeated_cache(alltypes, con):
    non_cached_table = alltypes.mutate(
        test_column=ibis.literal("calculation"), other_column=ibis.literal("big calc 2")
    )
    cached_table = non_cached_table.cache()
    nested_cached_table = cached_table.cache()
    name = cached_table.op().name

    assert not nested_cached_table.to_pandas().empty

    del nested_cached_table, cached_table

    assert name not in con.list_tables()


@mark.notimpl(["flink", "impala", "trino", "druid"])
@mark.notimpl(["exasol"], reason="Exasol does not support temporary tables")
@pytest.mark.never(
    ["risingwave"],
    raises=com.UnsupportedOperationError,
    reason="Feature is not yet implemented: CREATE TEMPORARY TABLE",
)
def test_persist_expression_release(con, alltypes):
    non_cached_table = alltypes.mutate(
        test_column=ibis.literal("calculation"), other_column=ibis.literal("big calc 3")
    )
    cached_table = non_cached_table.cache()
    cached_table.release()

    assert non_cached_table.op() not in con._cache_op_to_entry

    # a second release does not hurt
    cached_table.release()

    with pytest.raises(Exception, match=cached_table.op().name):
        cached_table.execute()



from __future__ import annotations

import contextlib
import importlib
import importlib.metadata
import itertools
from functools import cache
from pathlib import Path
from typing import TYPE_CHECKING, Any

import _pytest
import pytest
from packaging.requirements import Requirement

import ibis
from ibis import util
from ibis.backends import (
    CanCreateCatalog,
    CanCreateDatabase,
    _get_backend_names,
)
from ibis.conftest import WINDOWS
from ibis.util import promote_tuple

if TYPE_CHECKING:
    from ibis.backends import BaseBackend
    from ibis.backends.tests.base import BackendTest


TEST_TABLES = {
    "functional_alltypes": ibis.schema(
        {
            "id": "int32",
            "bool_col": "boolean",
            "tinyint_col": "int8",
            "smallint_col": "int16",
            "int_col": "int32",
            "bigint_col": "int64",
            "float_col": "float32",
            "double_col": "float64",
            "date_string_col": "string",
            "string_col": "string",
            "timestamp_col": "timestamp",
            "year": "int32",
            "month": "int32",
        }
    ),
    "diamonds": ibis.schema(
        {
            "carat": "float64",
            "cut": "string",
            "color": "string",
            "clarity": "string",
            "depth": "float64",
            "table": "float64",
            "price": "int64",
            "x": "float64",
            "y": "float64",
            "z": "float64",
        }
    ),
    "batting": ibis.schema(
        {
            "playerID": "string",
            "yearID": "int64",
            "stint": "int64",
            "teamID": "string",
            "lgID": "string",
            "G": "int64",
            "AB": "int64",
            "R": "int64",
            "H": "int64",
            "X2B": "int64",
            "X3B": "int64",
            "HR": "int64",
            "RBI": "int64",
            "SB": "int64",
            "CS": "int64",
            "BB": "int64",
            "SO": "int64",
            "IBB": "int64",
            "HBP": "int64",
            "SH": "int64",
            "SF": "int64",
            "GIDP": "int64",
        }
    ),
    "awards_players": ibis.schema(
        {
            "playerID": "string",
            "awardID": "string",
            "yearID": "int64",
            "lgID": "string",
            "tie": "string",
            "notes": "string",
        }
    ),
    "astronauts": ibis.schema(
        {
            "id": "int64",
            "number": "int64",
            "nationwide_number": "int64",
            "name": "string",
            "original_name": "string",
            "sex": "string",
            "year_of_birth": "int64",
            "nationality": "string",
            "military_civilian": "string",
            "selection": "string",
            "year_of_selection": "int64",
            "mission_number": "int64",
            "total_number_of_missions": "int64",
            "occupation": "string",
            "year_of_mission": "int64",
            "mission_title": "string",
            "ascend_shuttle": "string",
            "in_orbit": "string",
            "descend_shuttle": "string",
            "hours_mission": "float64",
            "total_hrs_sum": "float64",
            "field21": "int64",
            "eva_hrs_mission": "float64",
            "total_eva_hrs": "float64",
        }
    ),
}

# We want to check for exceptions in xfail tests for two reasons:
# * xfail tests without exception checking may hide problems.
#   For example, the implementation may work, but we may have an error in the test that needs to be fixed.
# * facilitates code reviews
# For now, many of our tests don't do this, and we're working to change this situation
# by improving all tests file by file. All files that have already been improved are
# added to this list to prevent regression.
FILES_WITH_STRICT_EXCEPTION_CHECK = [
    "ibis/backends/tests/test_api.py",
    "ibis/backends/tests/test_array.py",
    "ibis/backends/tests/test_aggregation.py",
    "ibis/backends/tests/test_binary.py",
    "ibis/backends/tests/test_numeric.py",
    "ibis/backends/tests/test_column.py",
    "ibis/backends/tests/test_string.py",
    "ibis/backends/tests/test_temporal.py",
    "ibis/backends/tests/test_uuid.py",
    "ibis/backends/tests/test_window.py",
]

ALL_BACKENDS = set(_get_backend_names())


@pytest.fixture(scope="session")
def data_dir() -> Path:
    """Return the test data directory.

    Returns
    -------
    Path
        Test data directory
    """
    root = Path(__file__).absolute().parents[2]

    return root / "ci" / "ibis-testing-data"


def _get_backend_conf(backend_str: str):
    """Convert a backend string to the test class for the backend."""
    conftest = importlib.import_module(f"ibis.backends.{backend_str}.tests.conftest")
    return conftest.TestConf


def _get_backend_from_parts(parts: tuple[str, ...]) -> str | None:
    """Return the backend part of a test file's path parts.

    Examples
    --------
    >>> _get_backend_from_parts(("/", "ibis", "backends", "sqlite", "tests"))
    "sqlite"
    """
    try:
        index = parts.index("backends")
    except ValueError:
        return None
    else:
        return parts[index + 1]


def pytest_ignore_collect(collection_path, config):
    # get the backend path part
    backend = _get_backend_from_parts(collection_path.parts)
    if backend is None or backend not in _get_backend_names():
        return False

    # we evaluate the marker early so that we don't trigger
    # an import of conftest files for the backend, which will
    # import the backend and thus require dependencies that may not
    # exist
    #
    # alternatives include littering library code with pytest.importorskips
    # and moving all imports close to their use site
    #
    # the latter isn't tenable for backends that use multiple dispatch
    # since the rules are executed at import time
    mark_expr = config.getoption("-m")
    # we can't let the empty string pass through, since `'' in s` is `True` for
    # any `s`; if no marker was passed don't ignore the collection of `path`
    if not mark_expr:
        return False
    expr = _pytest.mark.expression.Expression.compile(mark_expr)
    # we check the "backend" marker as well since if that's passed
    # any file matching a backend should be skipped
    keep = expr.evaluate(lambda s: s in (backend, "backend"))
    return not keep


def pytest_collection_modifyitems(session, config, items):
    all_backends = _get_backend_names()
    additional_markers = []

    unrecognized_backends = set()
    for item in items:
        # Yell loudly if unrecognized backend in notimpl, notyet or never
        for name in ("notimpl", "notyet", "never"):
            for mark in item.iter_markers(name=name):
                if backend := set(util.promote_list(mark.args[0])).difference(
                    ALL_BACKENDS
                ):
                    unrecognized_backends.add(
                        f"""Unrecognize backend(s) {backend} passed to {name} marker in
{item.path}::{item.originalname}"""
                    )

        # add the backend marker to any tests are inside "ibis/backends"
        parts = item.path.parts
        backend = _get_backend_from_parts(parts)
        if backend is not None and backend in all_backends:
            item.add_marker(getattr(pytest.mark, backend))
            item.add_marker(pytest.mark.backend)
        elif "backends" not in parts and not tuple(
            itertools.chain(
                *(item.iter_markers(name=name) for name in all_backends),
                item.iter_markers(name="backend"),
            )
        ):
            # anything else is a "core" test and is run by default
            if not any(item.iter_markers(name="benchmark")):
                item.add_marker(pytest.mark.core)

    if unrecognized_backends:
        raise pytest.PytestCollectionWarning("\n" + "\n".join(unrecognized_backends))

    for item, markers in additional_markers:
        for marker in markers:
            item.add_marker(marker)


@cache
def _get_backends_to_test(
    keep: tuple[str, ...] = (),
    discard: tuple[str, ...] = (),
) -> list[Any]:
    """Get a list of `TestConf` classes of the backends to test."""
    backends = _get_backend_names()

    if discard:
        backends = backends.difference(discard)

    if keep:
        backends = backends.intersection(keep)

    return [
        pytest.param(
            backend,
            marks=[getattr(pytest.mark, backend), pytest.mark.backend],
            id=backend,
        )
        for backend in sorted(backends)
    ]


def pytest_runtest_call(item):
    """Dynamically add various custom markers."""
    backend = [
        getattr(backend, "name", lambda backend=backend: backend)()
        for key, backend in item.funcargs.items()
        if key.endswith(("backend", "backend_name", "backend_no_data"))
    ]
    if len(backend) > 1:
        raise ValueError(
            f"test {item.originalname} was supplied with multiple backend "
            f"objects simultaneously. This is likely due to a leaky fixture."
            f"Backends passed: {(back.name() for back in backend)}"
        )
    if not backend:
        # Check item path to see if test is in backend-specific folder
        backend = set(_get_backend_names()).intersection(item.path.parts)
    if not backend:
        # Check if this is one of the uninstantiated backend class fixture
        # used for signature checking
        backend = [
            backend.name
            for key, backend in item.funcargs.items()
            if key.endswith("backend_cls")
        ]

    if not backend:
        return

    backend = next(iter(backend))

    def _filter_none_from_raises(kwargs):
        # Filter out any None values from kwargs['raises']
        # to cover any missing backend error types as defined in ibis/backends/tests/errors.py
        if (raises := kwargs.get("raises")) is not None:
            raises = tuple(filter(None, promote_tuple(raises)))
            if raises:
                kwargs["raises"] = raises
            else:
                # if filtering removes all of the values of raises pop the
                # argument otherwise it gets passed as an empty tuple and this
                # messes up xfail
                kwargs.pop("raises")
        return kwargs

    # Ibis hasn't exposed existing functionality
    # This xfails so that you know when it starts to pass
    for marker in item.iter_markers(name="notimpl"):
        if backend in marker.args[0]:
            if (
                item.location[0] in FILES_WITH_STRICT_EXCEPTION_CHECK
                and "raises" not in marker.kwargs.keys()
            ):
                raise ValueError("notimpl requires a raises")
            kwargs = marker.kwargs.copy()
            kwargs.setdefault("reason", f"Feature not yet exposed in {backend}")
            kwargs = _filter_none_from_raises(kwargs)
            item.add_marker(pytest.mark.xfail(**kwargs))

    # Functionality is unavailable upstream (but could be)
    # This xfails so that you know when it starts to pass
    for marker in item.iter_markers(name="notyet"):
        if backend in marker.args[0]:
            if (
                item.location[0] in FILES_WITH_STRICT_EXCEPTION_CHECK
                and "raises" not in marker.kwargs.keys()
            ):
                raise ValueError("notyet requires a raises")

            kwargs = marker.kwargs.copy()
            kwargs.setdefault("reason", f"Feature not available upstream for {backend}")
            kwargs = _filter_none_from_raises(kwargs)
            item.add_marker(pytest.mark.xfail(**kwargs))

    for marker in item.iter_markers(name="never"):
        if backend in marker.args[0]:
            if "reason" not in marker.kwargs.keys():
                raise ValueError("never requires a reason")
            kwargs = marker.kwargs.copy()
            kwargs = _filter_none_from_raises(kwargs)
            item.add_marker(pytest.mark.xfail(**kwargs))

    for marker in item.iter_markers(name="xfail_version"):
        kwargs = marker.kwargs.copy()
        kwargs = _filter_none_from_raises(kwargs)
        if backend not in kwargs:
            continue

        provided_reason = kwargs.pop("reason", None)
        specs = kwargs.pop(backend)
        failing_specs = []
        for spec in specs:
            req = Requirement(spec)
            if req.specifier.contains(
                importlib.import_module(req.name).__version__
            ) and ((not req.marker) or req.marker.evaluate()):
                failing_specs.append(spec)
        reason = f"{backend} backend test fails with {backend}{specs}"
        if provided_reason is not None:
            reason += f"; {provided_reason}"
        if failing_specs:
            item.add_marker(pytest.mark.xfail(reason=reason, **kwargs))


def _get_backend_cls(backend_str: str):
    """Convert a backend string to the test class for the backend."""
    backend_mod = importlib.import_module(f"ibis.backends.{backend_str}")
    return backend_mod.Backend


@pytest.fixture(params=_get_backends_to_test(), scope="session")
def backend_cls(request) -> BaseBackend:
    """Return the uninstantiated backend class, unconnected.

    This is used for signature checking and nothing should be executed."""

    cls = _get_backend_cls(request.param)
    return cls


@pytest.fixture(params=_get_backends_to_test(), scope="session")
def backend(request, data_dir, tmp_path_factory, worker_id) -> BackendTest:
    """Return an instance of BackendTest, loaded with data."""

    cls = _get_backend_conf(request.param)
    return cls.load_data(data_dir, tmp_path_factory, worker_id)


@pytest.fixture(scope="session")
def con(backend):
    """Instance of a backend client."""
    return backend.connection


@pytest.fixture(params=_get_backends_to_test(), scope="session")
def backend_no_data(request, data_dir, tmp_path_factory, worker_id):
    """Return an instance of BackendTest, with no data loaded."""
    cls = _get_backend_conf(request.param)
    return cls(data_dir=data_dir, tmpdir=tmp_path_factory, worker_id=worker_id)


@pytest.fixture(scope="session")
def con_no_data(backend_no_data):
    """Return an Ibis backend instance, with no data loaded."""
    return backend_no_data.connection


@pytest.fixture(scope="session")
def con_create_catalog(con):
    if isinstance(con, CanCreateCatalog):
        return con
    else:
        pytest.skip(f"{con.name} backend cannot create databases")


@pytest.fixture(scope="session")
def con_create_database(con):
    if isinstance(con, CanCreateDatabase):
        return con
    else:
        pytest.skip(f"{con.name} backend cannot create schemas")


@pytest.fixture(scope="session")
def con_create_catalog_database(con):
    if isinstance(con, CanCreateCatalog) and isinstance(con, CanCreateDatabase):
        return con
    else:
        pytest.skip(f"{con.name} backend cannot create both database and schemas")


def _setup_backend(request, data_dir, tmp_path_factory, worker_id):
    if (backend := request.param) == "duckdb" and WINDOWS:
        pytest.xfail(
            "windows prevents two connections to the same duckdb file "
            "even in the same process"
        )
    else:
        cls = _get_backend_conf(backend)
        return cls.load_data(data_dir, tmp_path_factory, worker_id)


@pytest.fixture(
    params=_get_backends_to_test(),
    scope="session",
)
def ddl_backend(request, data_dir, tmp_path_factory, worker_id):
    """Set up the backends that are SQL-based."""
    return _setup_backend(request, data_dir, tmp_path_factory, worker_id)


@pytest.fixture(scope="session")
def ddl_con(ddl_backend):
    """Instance of Client, already connected to the db (if applies)."""
    return ddl_backend.connection


@pytest.fixture(
    params=_get_backends_to_test(keep=("pyspark",)),
    scope="session",
)
def udf_backend(request, data_dir, tmp_path_factory, worker_id):
    """Runs the UDF-supporting backends."""
    cls = _get_backend_conf(request.param)
    return cls.load_data(data_dir, tmp_path_factory, worker_id)


@pytest.fixture(scope="session")
def alltypes(backend):
    return backend.functional_alltypes


@pytest.fixture(scope="session")
def json_t(backend):
    return backend.json_t


@pytest.fixture(scope="session")
def struct(backend):
    return backend.struct


@pytest.fixture(scope="session")
def udf_alltypes(udf_backend):
    return udf_backend.functional_alltypes


@pytest.fixture(scope="session")
def batting(backend):
    return backend.batting


@pytest.fixture(scope="session")
def awards_players(backend):
    return backend.awards_players


@pytest.fixture(scope="session")
def df(alltypes):
    return alltypes.execute()


@pytest.fixture(scope="session")
def struct_df(struct):
    return struct.execute()


@pytest.fixture(scope="session")
def udf_df(udf_alltypes):
    return udf_alltypes.execute()


@pytest.fixture(scope="session")
def sorted_df(df):
    return df.sort_values("id").reset_index(drop=True)


@pytest.fixture(scope="session")
def batting_df(batting):
    return batting.execute(limit=None)


@pytest.fixture(scope="session")
def awards_players_df(awards_players):
    return awards_players.execute(limit=None)


@pytest.fixture
def temp_table(con):
    """Return a temporary table name.

    Parameters
    ----------
    con : ibis.backends.Client

    Yields
    ------
    name : string
        Random table name for a temporary usage.
    """
    name = util.gen_name("temp_table")
    yield name
    with contextlib.suppress(NotImplementedError):
        con.drop_table(name, force=True)


@pytest.fixture
def temp_table2(con):
    name = util.gen_name("temp_table2")
    yield name
    with contextlib.suppress(NotImplementedError):
        con.drop_table(name, force=True)


@pytest.fixture
def temp_table_orig(con, temp_table):
    name = f"{temp_table}_orig"
    yield name
    with contextlib.suppress(NotImplementedError):
        con.drop_table(name, force=True)


@pytest.fixture
def temp_view(ddl_con):
    """Return a temporary view name.

    Parameters
    ----------
    ddl_con : backend connection

    Yields
    ------
    name : string
        Random view name for a temporary usage.
    """
    name = util.gen_name("view")
    yield name
    with contextlib.suppress(NotImplementedError):
        ddl_con.drop_view(name, force=True)


@pytest.fixture
def assert_sql(con, snapshot):
    def checker(expr, file_name="out.sql"):
        sql = con.compile(expr, pretty=True)
        snapshot.assert_match(sql, file_name)

    return checker



"""The following was forked from the python-interface project:

* Copyright (c) 2016-2021, Scott Sanderson

Utilities for typed interfaces.
"""

# ruff: noqa: D205, D415, D400

from __future__ import annotations

from functools import partial
from inspect import Parameter, Signature
from itertools import starmap, takewhile, zip_longest


def valfilter(f, d):
    return {k: v for k, v in d.items() if f(v)}


def dzip(left, right):
    return {k: (left.get(k), right.get(k)) for k in left.keys() & right.keys()}


def complement(f):
    def not_f(*args, **kwargs):
        return not f(*args, **kwargs)

    return not_f


def compatible(
    impl_sig: Signature, iface_sig: Signature, check_annotations: bool = True
) -> bool:
    """Check whether ``impl_sig`` is compatible with ``iface_sig``.

    Parameters
    ----------
    impl_sig
        The signature of the implementation function.
    iface_sig
        The signature of the interface function.
    check_annotations
        Whether to also compare signature annotations (default) vs only parameter names.

    In general, an implementation is compatible with an interface if any valid
    way of passing parameters to the interface method is also valid for the
    implementation.

    Consequently, the following differences are allowed between the signature
    of an implementation method and the signature of its interface definition:

    1. An implementation may add new arguments to an interface iff:
       a. All new arguments have default values.
       b. All new arguments accepted positionally (i.e. all non-keyword-only
          arguments) occur after any arguments declared by the interface.
       c. Keyword-only arguments may be reordered by the implementation.

    2. For type-annotated interfaces, type annotations my differ as follows:
       a. Arguments to implementations of an interface may be annotated with
          a **superclass** of the type specified by the interface.
       b. The return type of an implementation may be annotated with a
          **subclass** of the type specified by the interface.
    """
    # Unwrap to get the underlying inspect.Signature objects.
    return all(
        [
            positionals_compatible(
                takewhile(is_positional, impl_sig.parameters.values()),
                takewhile(is_positional, iface_sig.parameters.values()),
                check_annotations=check_annotations,
            ),
            keywords_compatible(
                valfilter(complement(is_positional), impl_sig.parameters),
                valfilter(complement(is_positional), iface_sig.parameters),
                check_annotations=check_annotations,
            ),
        ]
    )


_POSITIONALS = frozenset([Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD])


def is_positional(arg):
    return arg.kind in _POSITIONALS


def has_default(arg):
    """Does ``arg`` provide a default?."""
    return arg.default is not Parameter.empty


def params_compatible(impl, iface, check_annotations=True):
    if impl is None:
        return False

    if iface is None:
        return has_default(impl)

    checks = (
        impl.name == iface.name
        and impl.kind == iface.kind
        and has_default(impl) == has_default(iface)
    )

    if check_annotations:
        checks = checks and annotations_compatible(impl, iface)

    return checks


def positionals_compatible(impl_positionals, iface_positionals, check_annotations=True):
    params_compat = partial(params_compatible, check_annotations=check_annotations)
    return all(
        starmap(
            params_compat,
            zip_longest(impl_positionals, iface_positionals),
        )
    )


def keywords_compatible(impl_keywords, iface_keywords, check_annotations=True):
    params_compat = partial(params_compatible, check_annotations=check_annotations)
    return all(starmap(params_compat, dzip(impl_keywords, iface_keywords).values()))


def annotations_compatible(impl, iface):
    """Check whether the type annotations of an implementation are compatible with
    the annotations of the interface it implements.
    """
    return impl.annotation == iface.annotation



from __future__ import annotations

import os
from operator import attrgetter, methodcaller

import numpy.testing as npt
import pandas.testing as tm
import pyarrow as pa
import pytest
from packaging.version import parse as vparse
from pytest import param

import ibis
from ibis.conftest import LINUX, MACOS, SANDBOXED, WINDOWS

duckdb = pytest.importorskip("duckdb")
gpd = pytest.importorskip("geopandas")
gtm = pytest.importorskip("geopandas.testing")
shapely = pytest.importorskip("shapely")

Point = shapely.Point


def test_geospatial_point(zones, zones_gdf):
    coord = zones.x_cent.point(zones.y_cent).name("coord")
    # this returns GeometryArray
    gp_coord = gpd.points_from_xy(zones_gdf.x_cent, zones_gdf.y_cent)

    npt.assert_array_equal(coord.to_pandas().values, gp_coord)


# this functions are not implemented in geopandas
@pytest.mark.parametrize("operation", ["as_text", "n_points"])
def test_geospatial_unary_snapshot(operation, assert_sql):
    t = ibis.table([("geom", "geometry")], name="t")
    expr = getattr(t.geom, operation)().name("tmp")
    assert_sql(expr)


def test_geospatial_dwithin(assert_sql):
    t = ibis.table([("geom", "geometry")], name="t")
    expr = t.geom.d_within(t.geom, 3.0).name("tmp")
    assert_sql(expr)


# geospatial unary functions that return a non-geometry series
# we can test using pd.testing (tm)
@pytest.mark.parametrize(
    ("op", "gp_op"),
    [
        param("area", "area", id="area"),
        param("is_valid", "is_valid", id="is_valid"),
        param(
            "geometry_type",
            "geom_type",
            id="geometry_type",
            marks=pytest.mark.xfail(raises=pa.lib.ArrowTypeError),
        ),
    ],
)
def test_geospatial_unary_tm(op, gp_op, zones, zones_gdf):
    expr = getattr(zones.geom, op)().name("tmp")
    gp_expr = getattr(zones_gdf.geometry, gp_op)

    tm.assert_series_equal(expr.to_pandas(), gp_expr, check_names=False)


@pytest.mark.parametrize("op", ["x", "y"], ids=["x_coord", "y_coord"])
def test_geospatial_xy(op, zones, zones_gdf):
    cen = zones.geom.centroid().name("centroid")
    gp_cen = zones_gdf.geometry.centroid

    expr = getattr(cen, op)().name("tmp")
    gp_expr = getattr(gp_cen, op)

    tm.assert_series_equal(expr.to_pandas(), gp_expr, check_names=False)


def test_geospatial_length(lines, lines_gdf):
    # note: ST_LENGTH returns 0 for the case of polygon
    # or multi polygon while pandas geopandas returns the perimeter.
    length = lines.geom.length().name("length")
    gp_length = lines_gdf.geometry.length

    tm.assert_series_equal(length.to_pandas(), gp_length, check_names=False)


# geospatial binary functions that return a non-geometry series
# we can test using pd.testing (tm)
@pytest.mark.parametrize(
    ("op", "gp_op"),
    [
        param("contains", "contains", id="contains"),
        param("geo_equals", "geom_equals", id="geo_eqs"),
        param("covers", "covers", id="covers"),
        param("covered_by", "covered_by", id="covered_by"),
        param("crosses", "crosses", id="crosses"),
        param("disjoint", "disjoint", id="disjoint"),
        param("distance", "distance", id="distance"),
        param("intersects", "intersects", id="intersects"),
        param("overlaps", "overlaps", id="overlaps"),
        param("touches", "touches", id="touches"),
        param("within", "within", id="within"),
    ],
)
def test_geospatial_binary_tm(op, gp_op, zones, zones_gdf):
    expr = getattr(zones.geom, op)(zones.geom).name("tmp")
    gp_func = getattr(zones_gdf.geometry, gp_op)(zones_gdf.geometry)

    tm.assert_series_equal(expr.to_pandas(), gp_func, check_names=False)


# geospatial unary functions that return a geometry series
# we can test using gpd.testing (gtm)
@pytest.mark.parametrize("op", ["centroid", "envelope"])
def test_geospatial_unary_gtm(op, zones, zones_gdf):
    expr = getattr(zones.geom, op)().name("tmp")
    gp_expr = getattr(zones_gdf.geometry, op)

    gtm.assert_geoseries_equal(expr.to_pandas(), gp_expr, check_crs=False)


# geospatial binary functions that return a geometry series
# we can test using gpd.testing (gtm)
@pytest.mark.parametrize("op", ["difference", "intersection", "union"])
def test_geospatial_binary_gtm(op, zones, zones_gdf):
    expr = getattr(zones.geom, op)(zones.geom).name("tmp")
    gp_func = getattr(zones_gdf.geometry, op)(zones_gdf.geometry)

    gtm.assert_geoseries_equal(expr.to_pandas(), gp_func, check_crs=False)


def test_geospatial_end_point(lines, lines_gdf):
    epoint = lines.geom.end_point().name("end_point")
    # geopandas does not have end_point this is a work around to get it
    gp_epoint = lines_gdf.geometry.boundary.explode(index_parts=True).xs(1, level=1)

    gtm.assert_geoseries_equal(epoint.to_pandas(), gp_epoint, check_crs=False)


def test_geospatial_start_point(lines, lines_gdf):
    spoint = lines.geom.start_point().name("start_point")
    # geopandas does not have start_point this is a work around to get it
    gp_spoint = lines_gdf.geometry.boundary.explode(index_parts=True).xs(0, level=1)

    gtm.assert_geoseries_equal(spoint.to_pandas(), gp_spoint, check_crs=False)


# this one takes a bit longer than the rest.
@pytest.mark.parametrize(
    "expected_func",
    [
        param(
            attrgetter("unary_union"),
            marks=pytest.mark.xfail(
                condition=vparse(gpd.__version__) >= vparse("1"),
                raises=Warning,
                reason="unary_union property is deprecated",
            ),
            id="version<1",
        ),
        param(
            methodcaller("union_all"),
            marks=pytest.mark.xfail(
                condition=(
                    vparse(duckdb.__version__) < vparse("1.1.1")
                    and (
                        vparse(gpd.__version__) < vparse("1")
                        or vparse(shapely.__version__) >= vparse("2.0.5")
                    )
                ),
                raises=(AttributeError, AssertionError),
                reason="union_all doesn't exist; shapely 2.0.5 results in a different value for union_all",
            ),
            id="version>=1",
        ),
    ],
)
def test_geospatial_unary_union(zones, zones_gdf, expected_func):
    unary_union = zones.geom.unary_union().name("unary_union")
    # this returns a shapely geometry object
    gp_unary_union = expected_func(zones_gdf.geometry)

    # using set_precision because https://github.com/duckdb/duckdb_spatial/issues/189
    assert shapely.equals(
        shapely.set_precision(unary_union.to_pandas(), grid_size=1e-7),
        shapely.set_precision(gp_unary_union, grid_size=1e-7),
    )


def test_geospatial_buffer_point(zones, zones_gdf):
    cen = zones.geom.centroid().name("centroid")
    gp_cen = zones_gdf.geometry.centroid

    buffer = cen.buffer(100.0)
    # geopandas resolution default is 16, while duckdb is 8.
    gp_buffer = gp_cen.buffer(100.0, resolution=8)

    gtm.assert_geoseries_equal(buffer.to_pandas(), gp_buffer, check_crs=False)


def test_geospatial_buffer(zones, zones_gdf):
    buffer = zones.geom.buffer(100.0)
    # geopandas resolution default is 16, while duckdb is 8.
    gp_buffer = zones_gdf.geometry.buffer(100.0, resolution=8)

    gtm.assert_geoseries_equal(buffer.to_pandas(), gp_buffer, check_crs=False)


# using a smaller dataset for time purposes
def test_geospatial_convert(geotable, gdf):
    # geotable is fabricated but let's say the
    # data is in CRS: EPSG:2263
    # let's transform to EPSG:4326 (latitude-longitude projection)
    geo_ll = geotable.geom.convert("EPSG:2263", "EPSG:4326")

    gdf.crs = "EPSG:2263"
    gdf_ll = gdf.geometry.to_crs(crs=4326)

    gtm.assert_geoseries_equal(
        geo_ll.to_pandas(), gdf_ll, check_less_precise=True, check_crs=False
    )


def test_geospatial_flip_coordinates(geotable):
    flipped = geotable.geom.flip_coordinates()

    # flipped coords
    point = shapely.Point(40, -100)
    line_string = shapely.LineString([[0, 0], [1, 1], [1, 2], [2, 2]])
    polygon = shapely.Polygon(((0, 0), (0, 1), (1, 1), (1, 0), (0, 0)))

    d = {
        "name": ["Point", "LineString", "Polygon"],
        "geometry": [point, line_string, polygon],
    }

    flipped_gdf = gpd.GeoDataFrame(d)

    gtm.assert_geoseries_equal(
        flipped.to_pandas(), flipped_gdf.geometry, check_crs=False
    )


def test_create_table_geospatial_types(geotable, con):
    name = ibis.util.gen_name("geotable")

    # con = ibis.get_backend(geotable)

    t = con.create_table(name, geotable, temp=True)

    assert t.op().name in con.list_tables()
    assert any(map(methodcaller("is_geospatial"), t.schema().values()))


# geo literals declaration
point = ibis.literal((1, 0), type="point").name("p")
point_geom = ibis.literal((1, 0), type="point:geometry").name("p")


@pytest.mark.parametrize("expr", [point, point_geom])
def test_literal_geospatial_explicit(con, expr, assert_sql):
    assert_sql(expr)


# test input data with shapely geometries
shp_point_0 = shapely.Point(0, 0)
shp_point_1 = shapely.Point(1, 1)
shp_point_2 = shapely.Point(2, 2)

shp_linestring_0 = shapely.LineString([shp_point_0, shp_point_1, shp_point_2])
shp_linestring_1 = shapely.LineString([shp_point_2, shp_point_1, shp_point_0])
shp_polygon_0 = shapely.Polygon(shp_linestring_0)
shp_multilinestring_0 = shapely.MultiLineString([shp_linestring_0, shp_linestring_1])
shp_multipoint_0 = shapely.MultiPoint([shp_point_0, shp_point_1, shp_point_2])
shp_multipolygon_0 = shapely.MultiPolygon([shp_polygon_0])


@pytest.mark.parametrize(
    ("shp", "expected"),
    [
        (shp_point_0, "(0 0)"),
        (shp_point_1, "(1 1)"),
        (shp_point_2, "(2 2)"),
        (shp_linestring_0, "(0 0, 1 1, 2 2)"),
        (shp_linestring_1, "(2 2, 1 1, 0 0)"),
        (shp_polygon_0, "((0 0, 1 1, 2 2, 0 0))"),
        (shp_multipolygon_0, "(((0 0, 1 1, 2 2, 0 0)))"),
        (shp_multilinestring_0, "((0 0, 1 1, 2 2), (2 2, 1 1, 0 0))"),
        param(
            shp_multipoint_0,
            "((0 0), (1 1), (2 2))",
            marks=pytest.mark.xfail(
                raises=AssertionError,
                reason="Bug-fix change in GEOS 3.12 see shapely issue #1992",
            ),
        ),
    ],
)
def test_literal_geospatial_inferred(con, shp, expected, assert_sql):
    expr = ibis.literal(shp).name("result")
    result = con.compile(expr)
    name = type(shp).__name__.upper()
    pair = f"{name} {expected}"
    assert pair in result
    assert_sql(expr)


@pytest.mark.skipif(
    (LINUX or MACOS) and SANDBOXED,
    reason="nix on linux cannot download duckdb extensions or data due to sandboxing",
)
def test_load_geo_example(con):
    pytest.importorskip("pins")

    t = ibis.examples.zones.fetch(backend=con)
    assert t.geom.type().is_geospatial()


# For the next two tests we really want to ensure that
# load_extension("spatial") hasn't been run yet, so we create a new connection
# instead of using the con fixture.


@pytest.fixture(scope="session")
def geo_line_lit():
    return ibis.literal(shapely.LineString([[0, 0], [1, 0], [1, 1]]), type="geometry")


def test_geo_unop_geo_literals(con, geo_line_lit):
    """GeoSpatialUnOp operation on a geospatial literal"""
    expr = geo_line_lit.length()
    assert con.execute(expr) == 2


def test_geo_binop_geo_literals(con, geo_line_lit):
    """GeoSpatialBinOp operation on a geospatial literal"""
    expr = geo_line_lit.distance(shapely.Point(0, 0))
    assert con.execute(expr) == 0


def test_cast_wkb_to_geo(con):
    t = con.table("geo_wkb")
    geo_expr = t.geometry.cast("geometry")
    assert geo_expr.type().is_geospatial()
    assert isinstance(con.execute(geo_expr), gpd.GeoSeries)


def test_load_spatial_casting(data_dir, tmp_path_factory):
    # this directory is necessary because of Windows extension downloads race
    # condition
    con = ibis.duckdb.connect(extension_directory=tmp_path_factory.mktemp("extensions"))
    t = con.read_parquet(data_dir / "parquet" / "geo_wkb.parquet")

    geo_expr = t.limit(1).geometry.cast("geometry")

    assert geo_expr.type().is_geospatial()
    assert isinstance(con.execute(geo_expr), gpd.GeoSeries)


def test_geom_from_string(con):
    value = ibis.literal("POINT (1 2)")
    assert value.type().is_string()

    expr = value.cast("geometry")
    result = con.execute(expr)
    assert result == shapely.from_wkt("POINT (1 2)")


def no_roundtrip(
    *,
    reason: str,
    raises: type[Exception] | tuple[type[Exception], ...] = (
        duckdb.IOException,
        duckdb.NotImplementedException,
        duckdb.PermissionException,
    ),
):
    """Mark a test as expected to fail due to a reader/writer issue."""
    return pytest.mark.xfail(raises=raises, reason=reason)


@pytest.mark.parametrize(
    ("driver", "canonical_extension", "kwargs", "preproc"),
    [
        param("ESRI Shapefile", "shp", {}, None, id="shapefile"),
        param("MapInfo File", "mif", {}, None, id="mapinfo"),
        param(
            "S57",
            None,
            {},
            None,
            marks=no_roundtrip(
                reason="GDAL Error (1): Unable to load s57objectclasses.csv.  Unable to continue.",
            ),
            id="s57",
        ),
        param(
            "CSV",
            None,
            {"layer_creation_options": {"geometry_name": "geom", "geometry": "as_wkt"}},
            lambda t: t.mutate(geom=t.geom.as_text()),
            id="csv",
        ),
        ("GML", None, {}, None),
        param(
            "GPX",
            None,
            {},
            None,
            marks=no_roundtrip(reason="no geometry type specified"),
            id="gpx",
        ),
        param("KML", None, {}, None, id="kml"),
        param("GeoJSON", None, {}, None, id="geojson"),
        param("GeoJSONSeq", None, {}, None, id="geojsonseq"),
        param("OGR_GMT", "gmt", {}, None, id="gmt"),
        param("GPKG", None, {}, None, id="gpkg"),
        param("SQLite", None, {}, None, id="sqlite"),
        param(
            "WAsP",
            "map",
            {},
            None,
            marks=[
                no_roundtrip(reason="only linestrings are allowed"),
                pytest.mark.skipif(condition=WINDOWS, reason="hard crash on windows"),
            ],
            id="wasp",
        ),
        param(
            "OpenFileGDB",
            "gdb",
            {
                "layer_creation_options": {"geometry_name": "geom"},
                "geometry_type": "point",
            },
            None,
            id="gdb",
        ),
        param("FlatGeobuf", "fgb", {}, None, id="flatgeobuf"),
        param(
            "Geoconcept",
            "gxt",
            {"srs": "4326"},
            None,
            id="geoconcept",
            marks=no_roundtrip(reason="not entirely sure"),
        ),
        param(
            "PGDUMP",
            None,
            {"layer_creation_options": {"geometry_name": "geom"}},
            None,
            marks=no_roundtrip(reason="can only be read by postgres"),
            id="pgdump",
        ),
        param(
            "GPSBabel",
            None,
            {},
            None,
            id="gpsbabel",
            marks=no_roundtrip(
                reason="duckdb can't write this because it doesn't expose GDAL's dataset creation options"
            ),
        ),
        param(
            "ODS",
            "ods",
            {},
            lambda t: t.mutate(geom=t.geom.as_text()),
            marks=pytest.mark.xfail(
                condition=WINDOWS,
                raises=duckdb.IOException,
                reason="not working on windows",
            ),
            id="ods",
        ),
        param("XLSX", "xlsx", {}, lambda t: t.mutate(geom=t.geom.as_text()), id="xlsx"),
        param(
            "Selafin",
            None,
            {},
            None,
            marks=no_roundtrip(reason="duckdb wkb doesn't preserve the geometry type"),
            id="selafin",
        ),
        param("JML", None, {}, None, id="jml"),
        param("VDV", None, {}, lambda t: t.mutate(geom=t.geom.as_text()), id="vdv"),
        param(
            "MVT",
            "pbf",
            {},
            None,
            marks=no_roundtrip(reason="can't read the written file"),
            id="mvt",
        ),
        param("MapML", None, {}, None, id="mapml"),
        param(
            "PMTiles",
            None,
            {},
            None,
            id="pmtiles",
            marks=no_roundtrip(reason="row counts differ", raises=AssertionError),
        ),
        param("JSONFG", None, {}, None, id="jsonfg"),
    ],
)
def test_to_geo(con, driver, canonical_extension, kwargs, preproc, tmp_path):
    data = ibis.memtable({"x": [1, 3], "y": [2, 4]}).mutate(
        geom=lambda t: t.x.point(t.y)
    )

    if preproc is not None:
        data = preproc(data)

    ext = canonical_extension or driver.replace(" ", "_").lower()
    out = tmp_path / f"outfile.{ext}"

    con.to_geo(data, path=out, format=driver, **kwargs)
    dread = con.read_geo(out)

    assert dread.count().execute() == 2


GDAL_DATA = os.environ.get("GDAL_DATA")


@pytest.mark.parametrize(
    "driver",
    [
        param(
            "DGN",
            marks=pytest.mark.xfail(
                condition=GDAL_DATA is None,
                raises=duckdb.IOException,
                reason="GDAL_DATA not set",
            ),
        ),
        param(
            "DXF",
            marks=pytest.mark.xfail(
                condition=GDAL_DATA is None,
                raises=duckdb.IOException,
                reason="GDAL_DATA not set",
            ),
        ),
        "GEORSS",
    ],
)
def test_to_geo_geom_only(con, driver, tmp_path):
    data = ibis.memtable({"x": [1, 3], "y": [2, 4]}).select(
        geom=lambda t: t.x.point(t.y)
    )

    ext = driver.replace(" ", "_").lower()
    out = tmp_path / f"outfile.{ext}"

    con.to_geo(data, path=out, format=driver)
    dread = con.read_geo(out)

    assert dread.count().execute() == 2


def test_cache_geometry(con, monkeypatch):
    # ibis issue #10324

    # monkeypatching is necessary to ensure the correct backend is used for
    # caching
    monkeypatch.setattr(ibis.options, "default_backend", con)
    data = ibis.memtable({"x": [1], "y": [2]})
    data = data.select(geom=data.x.point(data.y)).cache()
    result = data.execute()
    assert result.at[0, "geom"] == shapely.Point(1, 2)
